__version__ = '0.1.0'
__author__ = 'Marcel Rennig and Daan Speth'
__author_email__ = ['daan.speth@univie.ac.at', 'marcel.rennig@univie.ac.at']
__copyright__ = 'Copyright 2024'
__credits__ = ['Daan Speth', 'Marcel Rennig']
__description__ = 'toolkit for working with amino acid sequences obtained from GlobDB genomes'
__license__ = 'GPL3'
__maintainer__ = 'Marcel Rennig and Daan Speth'
__maintainer_email__ = 'daan.speth@univie.ac.at'
__name__ = 'aastk'
__python_requires__ = '>=3.13'
__status__ = 'Production/Stable'
__title__ = 'aastk'
__url__ = 'https://github.com/dspeth/aastk'
